/*****************************************************************************

        AvsFilterMaxDif16.cpp
        Author: Laurent de Soras, 2011

--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*Tab=3***********************************************************************/



#if defined (_MSC_VER)
	#pragma warning (1 : 4130 4223 4705 4706)
	#pragma warning (4 : 4355 4786 4800)
#endif



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	"AvsFilterMaxDif16.h"

#include	<cassert>



/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



AvsFilterMaxDif16::AvsFilterMaxDif16 (::IScriptEnvironment *env_ptr, ::PClip src1_clip_sptr, ::PClip src2_clip_sptr, ::PClip ref_clip_sptr, int y, int u, int v)
:	GenericVideoFilter (src1_clip_sptr)
,	_src1_clip_sptr (src1_clip_sptr)
,	_src2_clip_sptr (src2_clip_sptr)
,	_ref_clip_sptr (ref_clip_sptr)
,	_plane_proc (vi, *this)
{
	// -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
	// Parameter check

	if (! vi.IsPlanar ())
	{
		env_ptr->ThrowError ("Dither_maxdif16: input must be planar.");
	}

	PlaneProcessor::check_same_format (env_ptr, vi, src2_clip_sptr, "Dither_maxdif16", "src2");
	PlaneProcessor::check_same_format (env_ptr, vi, ref_clip_sptr, "Dither_maxdif16", "ref");

	if (! PlaneProcessor::check_stack16_height (vi))
	{
		env_ptr->ThrowError (
			"Dither_maxdif16: stacked MSB/LSB data must have\n"
			"a height multiple of twice the maximum subsampling."
		);
	}

	_plane_proc.set_proc_mode (y, u, v);
	_plane_proc.set_clip_info (0, PlaneProcessor::ClipType_STACKED_16);
	_plane_proc.set_clip_info (1, PlaneProcessor::ClipType_STACKED_16);
	_plane_proc.set_clip_info (2, PlaneProcessor::ClipType_STACKED_16);
	_plane_proc.set_clip_info (3, PlaneProcessor::ClipType_STACKED_16);
}



::PVideoFrame __stdcall	AvsFilterMaxDif16::GetFrame (int n, ::IScriptEnvironment *env_ptr)
{
	::PVideoFrame	dst_sptr = env_ptr->NewVideoFrame (vi);

	_plane_proc.process_frame (
		dst_sptr, n, *env_ptr,
		&_src1_clip_sptr, &_src2_clip_sptr, &_ref_clip_sptr
	);

	return (dst_sptr);
}



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



void	AvsFilterMaxDif16::do_process_plane (::PVideoFrame &dst_sptr, int n, ::IScriptEnvironment &env, int plane_index, int plane_id)
{
	::PVideoFrame	src1_sptr = _src1_clip_sptr->GetFrame (n, &env);
	::PVideoFrame	src2_sptr = _src2_clip_sptr->GetFrame (n, &env);
	::PVideoFrame	ref_sptr  = _ref_clip_sptr->GetFrame (n, &env);

	const int		w            = _plane_proc.get_width (dst_sptr, plane_id);
	const int		h            = _plane_proc.get_height16 (dst_sptr, plane_id);
	uint8_t *		dst_msb_ptr  = dst_sptr->GetWritePtr (plane_id);
	const int		stride_dst   = dst_sptr->GetPitch (plane_id);

	const uint8_t*	src1_msb_ptr = src1_sptr->GetReadPtr (plane_id);
	const int		stride_src1  = src1_sptr->GetPitch (plane_id);
	const uint8_t*	src2_msb_ptr = src2_sptr->GetReadPtr (plane_id);
	const int		stride_src2  = src2_sptr->GetPitch (plane_id);
	const uint8_t*	ref_msb_ptr  = ref_sptr->GetReadPtr (plane_id);
	const int		stride_ref   = ref_sptr->GetPitch (plane_id);

	TaskDataGlobal		tdg;
	tdg._this_ptr    = this;
	tdg._w           = w;
	tdg._dst_msb_ptr = dst_msb_ptr;
	tdg._dst_lsb_ptr = dst_msb_ptr  + h * stride_dst;
	tdg._sr1_msb_ptr = src1_msb_ptr;
	tdg._sr1_lsb_ptr = src1_msb_ptr + h * stride_src1;
	tdg._sr2_msb_ptr = src2_msb_ptr;
	tdg._sr2_lsb_ptr = src2_msb_ptr + h * stride_src2;
	tdg._ref_msb_ptr = ref_msb_ptr;
	tdg._ref_lsb_ptr = ref_msb_ptr  + h * stride_ref;
	tdg._stride_dst  = stride_dst;
	tdg._stride_sr1  = stride_src1;
	tdg._stride_sr2  = stride_src2;
	tdg._stride_ref  = stride_ref;

	Slicer			slicer;
	slicer.start (h, tdg, &AvsFilterMaxDif16::process_subplane);
	slicer.wait ();
}



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



void	AvsFilterMaxDif16::process_subplane (Slicer::TaskData &td)
{
	assert (&td != 0);

	const TaskDataGlobal &	tdg = *(td._glob_data_ptr);
	uint8_t *		dst_msb_ptr = tdg._dst_msb_ptr + td._y_beg * tdg._stride_dst;
	uint8_t *		dst_lsb_ptr = tdg._dst_lsb_ptr + td._y_beg * tdg._stride_dst;
	const uint8_t*	sr1_msb_ptr = tdg._sr1_msb_ptr + td._y_beg * tdg._stride_sr1;
	const uint8_t*	sr1_lsb_ptr = tdg._sr1_lsb_ptr + td._y_beg * tdg._stride_sr1;
	const uint8_t*	sr2_msb_ptr = tdg._sr2_msb_ptr + td._y_beg * tdg._stride_sr2;
	const uint8_t*	sr2_lsb_ptr = tdg._sr2_lsb_ptr + td._y_beg * tdg._stride_sr2;
	const uint8_t*	ref_msb_ptr = tdg._ref_msb_ptr + td._y_beg * tdg._stride_ref;
	const uint8_t*	ref_lsb_ptr = tdg._ref_lsb_ptr + td._y_beg * tdg._stride_ref;

	for (int y = td._y_beg; y < td._y_end; ++y)
	{
		for (int x = 0; x < tdg._w; ++x)
		{
			const int		src1 = (sr1_msb_ptr [x] << 8) | sr1_lsb_ptr [x];
			const int		src2 = (sr2_msb_ptr [x] << 8) | sr2_lsb_ptr [x];
			const int		ref  = (ref_msb_ptr [x] << 8) | ref_lsb_ptr [x];

			int				val = src1;
			if (abs (src2 - ref) > abs (src1 - ref))
			{
				val = src2;
			}

			dst_msb_ptr [x] = val >> 8;
			dst_lsb_ptr [x] = val & 255;
		}

		dst_msb_ptr += tdg._stride_dst;
		dst_lsb_ptr += tdg._stride_dst;
		sr1_msb_ptr += tdg._stride_sr1;
		sr1_lsb_ptr += tdg._stride_sr1;
		sr2_msb_ptr += tdg._stride_sr2;
		sr2_lsb_ptr += tdg._stride_sr2;
		ref_msb_ptr += tdg._stride_ref;
		ref_lsb_ptr += tdg._stride_ref;
	}
}



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
