/*****************************************************************************

        AvsFilterResize16.h
        Author: Laurent de Soras, 2011

--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*Tab=3***********************************************************************/



#if ! defined (AvsFilterResize16_HEADER_INCLUDED)
#define	AvsFilterResize16_HEADER_INCLUDED

#if defined (_MSC_VER)
	#pragma once
	#pragma warning (4 : 4250)
#endif



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#define	WIN32_LEAN_AND_MEAN
#define	NOMINMAX
#define	NOGDI
#include <windows.h>
#include	"conc/Mutex.h"
#include	"AllocAlign.h"
#include	"avisynth.h"
#include	"ContFirInterface.h"
#include	"DiscreteFirInterface.h"
#include	"FilterResize.h"
#include	"KernelData.h"
#include	"PlaneProcCbInterface.h"
#include	"PlaneProcessor.h"
#include	"SharedPtr.h"
#include	"types.h"

#include	<emmintrin.h>

#include	<map>



class AvsFilterResize16
:	public GenericVideoFilter
,	public PlaneProcCbInterface
{

/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

public:

	enum OptBool
	{
		OptBool_UNDEF = -1,
		OptBool_FALSE = 0,
		OptBool_TRUE  = 1
	};

	typedef	AvsFilterResize16	ThisType;

	explicit			AvsFilterResize16 (::IScriptEnvironment *env_ptr, ::PClip src_clip_sptr, int dst_width, int dst_height, double win_x, double win_y, double win_w, double win_h, std::string kernel_fnc, double kernel_scale_h, double kernel_scale_v, int taps, bool a1_flag, double a1, bool a2_flag, double a2, bool a3_flag, double a3, int kovrspl, bool norm_flag, bool preserve_center_flag, std::string cplace, int y, int u, int v, std::string kernel_fnc_h, std::string kernel_fnc_v, double kernel_total_h, double kernel_total_v, bool invks_flag, OptBool invks_h, OptBool invks_v, int invks_taps);
	virtual			~AvsFilterResize16 () {}

	virtual ::PVideoFrame __stdcall
						GetFrame (int n, ::IScriptEnvironment *env_ptr);



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

protected:

	// PlaneProcCbInterface
	virtual void	do_process_plane (::PVideoFrame &dst_sptr, int n, ::IScriptEnvironment &env, int plane_index, int plane_id);



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

private:

	enum {			MAX_NBR_PLANES		= 3	};

	enum ChromaPlacement
	{
		ChromaPlacement_MPEG1 = 0,
		ChromaPlacement_MPEG2,
		ChromaPlacement_DV,

		ChromaPlacement_NBR_ELT
	};

	void				init_fnc (::IScriptEnvironment &env);

	FilterResize *	create_or_access_plane_filter (int plane_id, int src_width, int src_height, int dst_width, int dst_height, double win_x, double win_y, double win_w, double win_h, double center_pos_h, double center_pos_v);

	static ChromaPlacement
						conv_str_to_chroma_placement (::IScriptEnvironment &env, std::string cplace);
	static inline bool
						cumulate_flag (bool flag, OptBool flag2);

	::PClip			_src_clip_sptr;
	int				_src_width;
	int				_src_height;
	int				_dst_width;
	int				_dst_height;
	int				_invks_taps;
	double			_win_x;
	double			_win_y;
	double			_win_w;
	double			_win_h;
	std::string		_kernel_fnc_h;
	std::string		_kernel_fnc_v;
	double			_kernel_scale_h;			// Can be negative (forced scaling)
	double			_kernel_scale_v;			// Can be negative (forced scaling)
	double			_norm_val_h;
	double			_norm_val_v;
	bool				_norm_flag;
	bool				_invks_h_flag;
	bool				_invks_v_flag;
	bool				_preserve_center_flag;
	ChromaPlacement
						_chroma_placement;

	PlaneProcessor	_plane_proc;
	bool				_sse2_flag;
	conc::Mutex		_filter_mutex;				// To access _filter_aptr_map.
	std::map <int64_t, SharedPtr <FilterResize> >
						_filter_aptr_map;			// Created only on request. [(h<<32)+w]
	KernelData		_kernel_arr [FilterResize::Dir_NBR_ELT];



/*\\\ FORBIDDEN MEMBER FUNCTIONS \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

private:

						AvsFilterResize16 ();
						AvsFilterResize16 (const AvsFilterResize16 &other);
	AvsFilterResize16 &
						operator = (const AvsFilterResize16 &other);
	bool				operator == (const AvsFilterResize16 &other) const;
	bool				operator != (const AvsFilterResize16 &other) const;

};	// class AvsFilterResize16



//#include	"AvsFilterResize16.hpp"



#endif	// AvsFilterResize16_HEADER_INCLUDED



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
