/*****************************************************************************

        Scaler.h
        Author: Laurent de Soras, 2011

--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*Tab=3***********************************************************************/



#if ! defined (Scaler_HEADER_INCLUDED)
#define	Scaler_HEADER_INCLUDED

#if defined (_MSC_VER)
	#pragma once
	#pragma warning (4 : 4250)
#endif



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	"AllocAlign.h"
#include	"types.h"

#include	<emmintrin.h>

#include	<vector>



class ContFirInterface;

class Scaler
{

/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

public:

	enum {			SRC_ALIGN	= 8	};	// Pixels

	typedef	Scaler	ThisType;
	
	template <class T>
	class Ptr1
	{
	public:
		typedef	T *	Type;
		static __forceinline void
							jump (Type &ptr, int stride);
		static __forceinline bool
							check_ptr (const Type &ptr, int align = 1);
	};

	template <class T>
	class Ptr2
	{
	public:
		class Type
		{
		public:
			T *				_msb_ptr;
			T *				_lsb_ptr;
		};
		static __forceinline void
							jump (Type &ptr, int stride);
		static __forceinline bool
							check_ptr (const Type &ptr, int align = 1);
	};

	typedef	Ptr1 <float>			PtrFloat;
	typedef	Ptr1 <const float>	PtrFloatConst;
	typedef	Ptr2 <uint8_t>			PtrStack16;
	typedef	Ptr2 <const uint8_t>	PtrStack16Const;

	explicit			Scaler (int src_height, int dst_height, double win_top, double win_height, ContFirInterface &kernel_fnc, double kernel_scale, bool norm_flag, double norm_val, double center_pos, bool sse_flag);
	virtual			~Scaler () {}

	void				get_src_boundaries (int &y_src_beg, int &y_src_end, int y_dst_beg, int y_dst_end) const;
	int				get_fir_len () const;
	void				process_plane (PtrFloat::Type dst_ptr, PtrFloatConst::Type src_ptr, int dst_stride, int src_stride, int width, int y_dst_beg, int y_dst_end) const;
	void				process_plane (PtrFloat::Type dst_ptr, PtrStack16Const::Type src_ptr, int dst_stride, int src_stride, int width, int y_dst_beg, int y_dst_end) const;
	void				process_plane (PtrStack16::Type dst_ptr, PtrFloatConst::Type src_ptr, int dst_stride, int src_stride, int width, int y_dst_beg, int y_dst_end) const;
	void				process_plane (PtrStack16::Type dst_ptr, PtrStack16Const::Type src_ptr, int dst_stride, int src_stride, int width, int y_dst_beg, int y_dst_end) const;

	static void		eval_req_src_area (int &work_top, int &work_height, int src_height, int dst_height, double win_top, double win_height, ContFirInterface &kernel_fnc, double kernel_scale, double center_pos);
	static int		eval_lower_bound_of_dst_tile_height (int tile_height_src, int dst_height, double win_height, ContFirInterface &kernel_fnc, double kernel_scale, int src_height);
	static int		eval_lower_bound_of_src_tile_height (int tile_height_dst, int dst_height, double win_height, ContFirInterface &kernel_fnc, double kernel_scale, int src_height);



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

protected:



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

private:

	class BasicInfo
	{
	public:
							BasicInfo (int src_height, int dst_height, double win_top, double win_height, ContFirInterface &kernel_fnc, double kernel_scale, double center_pos);

		double			_src_step;
		double			_zc_size;
		double			_imp_step;
		double			_support;
		double			_src_pos;
		int				_fir_len;
	};

	class KernelInfo
	{
	public:
		int				_start_line;	// Y position of the first line of the kernel
		int				_coef_index;
		int				_kernel_size;
	};

	// Proxies

	template <class T> class ProxyRwCpp {};
	template <class T> class ProxyRwSse {};

	template <>
	class ProxyRwCpp <float>
	{
	public:
		typedef	PtrFloat			Ptr;
		typedef	PtrFloatConst	PtrConst;
		enum {			ALIGN	= 1	};
		static __forceinline void
							read (const PtrConst::Type &ptr, float &src0, float &src1);
		static __forceinline void
							write (const Ptr::Type &ptr, const float &src0, const float &src1);
	};

	template <>
	class ProxyRwCpp <uint8_t>
	{
	public:
		typedef	PtrStack16			Ptr;
		typedef	PtrStack16Const	PtrConst;
		enum {			ALIGN	= 1	};
		static __forceinline void
							read (const PtrConst::Type &ptr, float &src0, float &src1);
		static __forceinline void
							write (const Ptr::Type &ptr, const float &src0, const float &src1);
	};

	template <>
	class ProxyRwSse <float>
	{
	public:
		typedef	PtrFloat			Ptr;
		typedef	PtrFloatConst	PtrConst;
		enum {			ALIGN	= 16	};
		static __forceinline void
							read (const PtrConst::Type &ptr, __m128 &src0, __m128 &src1, const __m128i &zero);
		static __forceinline void
							write (const Ptr::Type &ptr, const __m128 &src0, const __m128 &src1, const __m128i &mask_lsb, const __m128i &sign_bit, const __m128 &offset);
		static __forceinline void
							write_partial (const Ptr::Type &ptr, const __m128 &src0, const __m128 &src1, const __m128i &mask_lsb, const __m128i &sign_bit, const __m128 &offset, const __m128i &mask_store);
	};

	template <>
	class ProxyRwSse <uint8_t>
	{
	public:
		typedef	PtrStack16			Ptr;
		typedef	PtrStack16Const	PtrConst;
		enum {			ALIGN	= 1	};
		static __forceinline void
							read (const PtrConst::Type &ptr, __m128 &src0, __m128 &src1, const __m128i &zero);
		static __forceinline void
							write (const Ptr::Type &ptr, const __m128 &src0, const __m128 &src1, const __m128i &mask_lsb, const __m128i &sign_bit, const __m128 &offset);
		static __forceinline void
							write_partial (const Ptr::Type &ptr, const __m128 &src0, const __m128 &src1, const __m128i &mask_lsb, const __m128i &sign_bit, const __m128 &offset, const __m128i &mask_store);
	};


	template <class DST, class SRC>
	void				process_plane_cpp (typename DST::Ptr::Type dst_ptr, typename SRC::PtrConst::Type src_ptr, int dst_stride, int src_stride, int width, int y_dst_beg, int y_dst_end) const;
	template <class DST, class SRC>
	void				process_plane_sse (typename DST::Ptr::Type dst_ptr, typename SRC::PtrConst::Type src_ptr, int dst_stride, int src_stride, int width, int y_dst_beg, int y_dst_end) const;
	template <class SRC>
	static __forceinline void
						process_vect_sse (__m128 &sum0, __m128 &sum1, int kernel_size, const float *coef_base_ptr, typename SRC::PtrConst::Type pix_ptr, const __m128i &zero, int src_stride);

	void				build_scale_data ();

	int				_src_height;
	int				_dst_height;
	double			_win_top;
	double			_win_height;
	double			_kernel_scale;
	ContFirInterface &
						_kernel_fnc;
	bool				_norm_flag;
	double			_norm_val;
	double			_center_pos;
	int				_fir_len;

	std::vector <KernelInfo>			// For each destination line
						_kernel_info_arr;
	std::vector <float, AllocAlign <float, 16> >	// All kernel coefs, for all lines.
						_coef_arr;

	void (ThisType::*
						_process_plane_f32_f32_ptr) (PtrFloat::Type dst_ptr, PtrFloatConst::Type src_ptr, int dst_stride, int src_stride, int width, int y_dst_beg, int y_dst_end) const;
	void (ThisType::*
						_process_plane_f32_s16_ptr) (PtrFloat::Type dst_ptr, PtrStack16Const::Type src_ptr, int dst_stride, int src_stride, int width, int y_dst_beg, int y_dst_end) const;
	void (ThisType::*
						_process_plane_s16_f32_ptr) (PtrStack16::Type dst_ptr, PtrFloatConst::Type src_ptr, int dst_stride, int src_stride, int width, int y_dst_beg, int y_dst_end) const;
	void (ThisType::*
						_process_plane_s16_s16_ptr) (PtrStack16::Type dst_ptr, PtrStack16Const::Type src_ptr, int dst_stride, int src_stride, int width, int y_dst_beg, int y_dst_end) const;



/*\\\ FORBIDDEN MEMBER FUNCTIONS \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

private:

						Scaler ();
						Scaler (const Scaler &other);
	Scaler &			operator = (const Scaler &other);
	bool				operator == (const Scaler &other) const;
	bool				operator != (const Scaler &other) const;

};	// class Scaler



#include	"Scaler.hpp"



#endif	// Scaler_HEADER_INCLUDED



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
